// 自动设置的参数
texture2D ScreenTexture : POSTEFFECTTEXTURE; // PostEffect 传递进来的第一个参数
sampler2D ScreenTextureSampler = sampler_state { // 采样器
	texture = <ScreenTexture>;
	AddressU  = BORDER;
	AddressV = BORDER;
	Filter = MIN_MAG_MIP_POINT; // 邻近采样，用 MIN_MAG_LINEAR_MIP_POINT 可以切换到双线性采样
};
float4 screen : SCREENSIZE; // 屏幕缓冲区大小（游戏窗口大小）
float4 viewport : VIEWPORT; // 可绘制区域，即视口

// PostEffect 传递的参数，通过 string binding 进行绑定
// lstg.Color 被解释为 float4
// number 被解释为 float
// string 被解释为 texture2D 纹理资源
float4 color <string binding = "color";> = float4(1.0f, 1.0f, 1.0f, 1.0f);
float number <string binding = "number";> = 1.0f;
texture2D UserTexture < string binding = "texture"; >;
sampler2D UserTextureSampler = sampler_state { // 采样器
	texture = <UserTexture>;
	AddressU = BORDER;
	AddressV = BORDER;
	Filter = MIN_MAG_LINEAR_MIP_POINT;
};

struct PS_IN {
	float4 position : POSITION;
	float2 uv : TEXCOORD;
};

struct PS_OUT {
	float4 color : COLOR;
};

PS_OUT PS(PS_IN input) {
	float2 screenSize = float2(screen.z - screen.x, screen.w - screen.y); // 屏幕大小
	float2 xy = input.uv * screenSize; // uv坐标对应的屏幕上的位置
	
	PS_OUT output;
	if (xy.x >= viewport.x && xy.x <= viewport.z && xy.y >= viewport.y && xy.x <= viewport.w) {
		float4 texcolor = tex2D(ScreenTextureSampler, input.uv); // 采样
		float4 texcolor2 = tex2D(UserTextureSampler, input.uv); // 采样
		output.color = color * texcolor * texcolor2; // 纹理颜色和指定颜色相乘
		output.color.a = number; // 修改不透明度
	}
	else {
		output.color.a = 0.0f; // 全透明，不显示这个像素
	}
	
	return output;
}

technique Main {
	pass MainPass {
		PixelShader = compile ps_3_0 PS();
	}
}
