----------------------------------------
--- LuaSTG 绘制教程 v2.0
--- 作者：Xiliusha
--- 邮箱：Xiliusha@outlook.com
----------------------------------------

----------------------------------------
--- 使用方法

--- 将 tutorial 这个文件夹放在 mod 文件夹内
---
--- 如果你用的是 LuaSTG Editor 编辑器，使用 code 节点
--- 写上 Include("mod/tutorial/tutorial.lua")
---
--- 如果你是打算只写代码，不使用 LuaSTG Editor 编辑器
--- 请在合适的地方写上 Include("mod/tutorial/tutorial.lua")
---
--- 直接在关卡内调用 tutorial.xxxx() 即可（xxxx替换成下面定义的任意函数）
--- 建议每个关卡内只调用一次，因为生成的绘制用的 object 不会自动删除

----------------------------------------
--- 教程代码库

---@class tutorial
tutorial = {}

--- 教程资源文件的存放路径
tutorial.path = "mod/tutorial/"

--- 自定义的事件集
---@class tutorial.host : lstg.GameObject
local tutorial_host = Class(object)
function tutorial_host:init(initf, framef, renderf)
	self.bound = false
	self.colli = false
	self.initf = initf
	self.framef = framef
	self.renderf = renderf
	initf(self)
end
function tutorial_host:frame()
	task.Do(self)
	self.framef(self)
end
function tutorial_host:render()
	self.renderf(self)
end

--- 用自定义的事件集创建游戏对象
---@param initf fun(self:tutorial.host)
---@param framef fun(self:tutorial.host)
---@param renderf fun(self:tutorial.host)
---@return tutorial.host
local function create_host(initf, framef, renderf)
	return lstg.New(tutorial_host,
			initf, framef, renderf)
end

local function infinite_wait()
	while true do
		task.Wait(1)
	end
end

----------------------------------------
--- 知识 可参与绘制的的资源

--- LuaSTG 中可参与绘制的资源有
--- 1、纹理（Texture）              枚举值 1，缩写 "tex"
--- 2、图片（Sprite）               枚举值 2，缩写 "img"
--- 3、动画（Animation）            枚举值 3，缩写 "ani"
--- 4、粒子特效（ParticleSystem）   枚举值 6，缩写 "psi"
--- 5、HGE 纹理字体（HGEFont）      枚举值 7，缩写 "fnt"
--- 6、TrueType 字体（TTFFont）     枚举值 8，缩写 "ttf"
--- 7、DirectX 9 Effect（Effect）  枚举值 9，缩写 "fx"
--- 后面的枚举值和缩写用于 lstg.EnumRes、lstg.CheckRes、lstg.RemoveResource

----------------------------------------
--- 知识 混合模式

--- LuaSTG 中可用的混合模式有
--- 1、"mul+alpha" 顶点颜色和纹理颜色相乘，上层和下层用不透明度通道进行颜色混合，也就是常说的透明度混合
--- 2、"mul+add"   顶点颜色和纹理颜色相乘，上层和下层颜色相加，高光子弹会用这个混合模式
--- 3、"mul+sub"   顶点颜色和纹理颜色相乘，上层颜色被下层颜色减去，可以做反色特效
--- 4、"mul+rev"   顶点颜色和纹理颜色相乘，下层颜色被上层颜色减去，做减色效果会用这个混合模式
--- 5、"add+alpha" 顶点颜色和纹理颜色相加，上层和下层用不透明度通道进行颜色混合，也就是常说的透明度混合
--- 6、"add+add"   顶点颜色和纹理颜色相加，上层和下层颜色相加，高光子弹会用这个混合模式
--- 7、"add+sub"   顶点颜色和纹理颜色相加，上层颜色被下层颜色减去，可以做反色特效
--- 8、"add+rev"   顶点颜色和纹理颜色相加，下层颜色被上层颜色减去，做减色效果会用这个混合模式
--- 9、"alpha+bal" 顶点颜色和纹理颜色相乘，下层的反色和下层进行颜色混合，上层的颜色通道充当不透明度通道，可以做不规则区域反色特效
--- 直接用 "" 等效于 "mul+alpha"，也就是默认的透明度混合
--- 第 9 个是 LuaSTG Ex Plus（也就是常说的 ex+）分支新增的

----------------------------------------
--- 知识 纹理uv坐标

--- 纹理 uv 坐标以左上角为原点，向右为 u 轴正方向，向下为 v 轴正方向
--- 一般 uv 坐标是归一化的，也就是从 0 到 1
--- 不过 LuaSTG 会以纹理宽度和纹理高度为 uv 坐标的范围
--- 比如一个 128x256 的纹理，右下角 uv 坐标为 u = 128, v = 256

----------------------------------------
--- 知识 对齐方式

--- LuaSTG 的对齐方式有垂直左中右和水平上中下共 9 种对齐方式
--- 传递参数时，要传递一个整数值，对齐方式的控制是按位设置的
--- Lua 无法直接编写位操作，所幸的是我们可以通过加法来模拟
--- 0 + 0 代表水平居左垂直居上对齐
--- 1 + 0 代表水平居中垂直居上对齐
--- 2 + 0 代表水平居右垂直居上对齐
--- 0 + 4 代表水平居左垂直居中对齐
--- 1 + 4 代表水平居中垂直居中对齐
--- 2 + 4 代表水平居右垂直居中对齐
--- 0 + 8 代表水平居左垂直居下对齐
--- 1 + 8 代表水平居中垂直居下对齐
--- 2 + 8 代表水平居右垂直居下对齐
--- 此外，如果只需要再矩形区域内绘制中文字符，还可以额外加上 16
--- 这样绘制的时候会自动换行，但是只支持中文字符，而且不处理标点符号等情况，也不支持英文单词等

----------------------------------------
--- 教程 加载可绘制的资源

local function load_resource()
	-- 加载纹理
	-- 参数（纹理名，纹理图片文件路径，开启Mipmap）
	lstg.LoadTexture("tex_pillar", tutorial.path .. "pillar.png", false)
	-- 加载图片
	-- 参数（图片名，纹理名，
	--         图片左上角u坐标，图片左上角v坐标，图片u轴上的宽度，图片v轴上的高度）
	lstg.LoadImage("img_pillar", "tex_pillar",
			0, 0, 32, 256)
	-- 加载动画（一组图片的集合）
	-- 参数（图片名，纹理名，
	--         单张图片左上角u坐标，单张图片左上角v坐标，单张图片u轴上的宽度，单张图片v轴上的高度，
	--         纵向图片数，横向图片数，动画间隔）
	lstg.LoadAnimation("ani_pillar", "tex_pillar",
			0, 0, 16, 64,
			2, 4, 4)
	-- 加载粒子特效
	-- 参数（粒子特效名，粒子特效定义文件路径，图片名）
	lstg.LoadPS("psi_death", tutorial.path .. "player_death_ef.psi", "img_pillar")
	-- 加载 HGE 纹理字体
	-- 需要注意的是 HGE 纹理字体定义文件要和配套的字体纹理图片放在同一个文件夹内
	-- 参数（HGE 纹理字体名，HGE 纹理字体定义文件路径，开启Mipmap）
	lstg.LoadFont("fnt_menu", tutorial.path .. "menu.fnt", false)
	-- 加载 TrueType 字体
	-- 字形像素宽度和字形像素高度其中一个要填 0，这样不容易加载失败
	-- 参数（TrueType 字体名，TrueType 字体定义文件路径，字形像素宽度，字形像素高度）
	lstg.LoadTTF("ttf_unifont", tutorial.path .. "unifont-13.0.02.ttf", 0, 64)
	-- 创建离屏渲染用的纹理
	-- 参数（纹理名）
	lstg.CreateRenderTarget("tex_buffer")
	-- 加载 DirectX 9 Effect
	-- 参数（后处理特效名，后处理特效源文件路径）
	lstg.LoadFX("fx_effect", tutorial.path .. "effect.fx")
end

----------------------------------------
--- 教程 直接用 object 来绘制

--- 在早期版本的 data 中 object 作为一个基础事件集，居然不能作为 New 的参数
--- 所幸的是 LuaSTG ex+0.81c 开始这个问题被修复了
--- object 可以说是最简单的事件集，其能做的事情也相当有限
--- 你可以修改它的 img 来绘制，但是无法很方便地修改它的绘制颜色
--- 支持绘制图片（Sprite）、动画（Animation）和粒子特效（ParticleSystem）

function tutorial.render_by_object()
	load_resource()
	
	local host =  lstg.New(object)
	
	-- DefaultRenderFunc 会用到的游戏对象属性
	host.img = "img_pillar"
	host.x = 0 -- x 坐标
	host.y = 0 -- y 坐标
	host.rot = 0 -- 朝向，角度制，逆时针为正方向
	host.hscale = 1 -- 横向缩放
	host.vscale = 1 -- 纵向缩放
	
	-- 其他属性
	host.omiga = 2 -- 自转
	
	infinite_wait()
end

----------------------------------------
--- 教程 改变顶点颜色和混合模式

--- 现在我们来用自定义的事件集，创建一个游戏对象， 动态修改它的绘制颜色
--- lstg.DefaultRenderFunc 是游戏对象的默认绘制方法
--- 它会读取游戏对象的 img, x, y, rot, hscale, vscale 来绘制
--- 支持绘制图片（Sprite）、动画（Animation）和粒子特效（ParticleSystem）
--- lstg.SetImgState 则可以修改 img 的顶点颜色和混合模式
--- 不过只能修改图片（Sprite）和动画（Animation）的顶点颜色和混合模式

function tutorial.change_color()
	load_resource()
	
	local host = create_host(
			function(self)
				self.layer = LAYER_BG + 50 -- 图层，可以修改
				
				-- DefaultRenderFunc 会用到的游戏对象属性
				self.img = "img_pillar"
				self.x = 0 -- x 坐标
				self.y = 0 -- y 坐标
				self.rot = 0 -- 朝向，角度制，逆时针为正方向
				self.hscale = 1 -- 横向缩放
				self.vscale = 1 -- 纵向缩放
				
				-- 其他属性
				self.omiga = 2 -- 自转
			end,
			function(self)
				-- 什么都不做
			end,
			function(self)
				local k = 0.5 + 0.5 * lstg.sin(self.timer * 8)
				-- 参数（游戏对象，混合模式，不透明度，红色，绿色，蓝色）
				lstg.SetImgState(self, "mul+alpha", 255, 255, k * 255, k * 255) -- 先设置
				-- 参数（游戏对象）
				lstg.DefaultRenderFunc(self) -- 然后绘制，这个是游戏对象的默认绘制函数
			end)
	
	infinite_wait()
end

----------------------------------------
--- 教程 改变顶点颜色和混合模式，另一种写法

--- 如果我们不使用 lstg.DefaultRenderFunc 和 lstg.SetImgState
--- 那我们可以通过 lstg.Render 和 lstg.SetImageState 来绘制
--- 不过，只能绘制图片（Sprite），但是这已经足够了
--- 从这里开始，我们会接触到 lstg.Color 这个颜色对象，用于 LuaSTG

function tutorial.change_color_other()
	load_resource()
	
	local host = create_host(
			function(self)
				self.layer = LAYER_BG + 50 -- 图层，可以修改
			end,
			function(self)
				-- 什么都不做
			end,
			function(self)
				local k = 0.5 + 0.5 * lstg.sin(self.timer * 8)
				-- 参数（不透明度，红色，绿色，蓝色）
				local color = lstg.Color(255, 255, k * 255, k * 255) -- 创建一个颜色对象
				-- 参数（图片名，混合模式，颜色）
				lstg.SetImageState("img_pillar", "mul+alpha", color) -- 先设置
				-- 参数（图片名，x坐标，y坐标，朝向，横向缩放，纵向缩放）
				lstg.Render("img_pillar", 0, 0, self.timer * 2, 1, 1) -- 然后绘制
			end)
	
	infinite_wait()
end

----------------------------------------
--- 教程 以颜色覆盖视口

--- 用某种颜色覆盖可绘制区域
--- 视口（Viewport），也就是可绘制区域，是计算机图形学中很重要的一个概念

function tutorial.clear_by_color()
	load_resource()
	
	local host = create_host(
			function(self)
				self.layer = LAYER_BG + 50 -- 图层，可以修改
			end,
			function(self)
				-- 什么都不做
			end,
			function(self)
				local k = 0.5 + 0.5 * lstg.sin(self.timer * 8)
				-- 参数（不透明度，红色，绿色，蓝色）
				-- 注意不透明度最好只填255，也就是完全不透明
				local color = lstg.Color(255, 255, k * 255, k * 255) -- 创建一个颜色对象
				-- 参数（颜色）
				lstg.RenderClear(color) -- 覆盖可绘制区域
			end)
	
	infinite_wait()
end

----------------------------------------
--- 教程 在一个矩形内绘制

--- 如果我们不打算用 lstg.Render 来绘制，还可以试试 lstg.RenderRect
--- 不过这个方法不能改变朝向

function tutorial.render_in_rect()
	load_resource()
	
	local host = create_host(
			function(self)
				self.layer = LAYER_BG + 50 -- 图层，可以修改
			end,
			function(self)
				-- 什么都不做
			end,
			function(self)
				local k = 0.5 + 0.5 * lstg.sin(self.timer * 8)
				-- 参数（不透明度，红色，绿色，蓝色）
				local color = lstg.Color(255, 255, k * 255, k * 255) -- 创建一个颜色对象
				-- 参数（图片名，混合模式，颜色）
				lstg.SetImageState("img_pillar", "mul+alpha", color) -- 先设置
				-- 参数（图片名，左边，右边，底边，顶边）
				lstg.RenderRect("img_pillar", -64, 128, -32, 16) -- 然后绘制
				lstg.RenderRect("img_pillar", -8, 16, -64, 128)
			end)
	
	infinite_wait()
end

----------------------------------------
--- 教程 指定顶点坐标绘制

--- lstg.Render4V 可以指定4个顶点的坐标，是一种复杂的方法
--- 目前，4个顶点坐标的z坐标填0.5就可以了，因为我们只需要绘制2D画面
--- 绘制的时候，需要注意顶点的顺序，一般来说图片左上角为第一个顶点
--- 图片右上角为第二个顶点，图片右下角为第三个顶点，图片左下角为第四个顶点
--- 需要注意的是，绘制的时候四个顶点必须形成空间同平面的平行四边形
--- 否则会出现失真的现象，在这个教程里面也会演示出来

function tutorial.render_by_4_position()
	load_resource()
	
	local host = create_host(
			function(self)
				self.layer = LAYER_BG + 50 -- 图层，可以修改
			end,
			function(self)
				-- 什么都不做
			end,
			function(self)
				local k = 0.5 + 0.5 * lstg.sin(self.timer * 8)
				-- 参数（不透明度，红色，绿色，蓝色）
				local color = lstg.Color(255, 255, k * 255, k * 255) -- 创建一个颜色对象
				-- 参数（图片名，混合模式，颜色）
				lstg.SetImageState("img_pillar", "mul+alpha", color) -- 先设置
				-- 参数（图片名，x坐标1，y坐标1，z坐标1，x坐标2，y坐标2，z坐标2，x坐标3，y坐标3，z坐标3，x坐标4，y坐标4，z坐标4）
				lstg.Render4V("img_pillar",
						-160, 160, 0.5,
						-120, 160, 0.5,
						-120, -160, 0.5,
						-160, -160, 0.5)
				-- 失真的情况
				lstg.Render4V("img_pillar",
						-100, 16, 0.5,
						160, 64, 0.5,
						50, -32, 0.5,
						-80, -16, 0.5)
			end)
	
	infinite_wait()
end

----------------------------------------
--- 教程 指定顶点绘制

--- lstg.RenderTexture 可以指定4个顶点，是最复杂的方法
--- 可以指定每个顶点的空间 xyz 坐标、纹理 uv 坐标和顶点颜色
--- 目前，4个顶点的z坐标填0.5就可以了，因为我们只需要绘制2D画面
--- 需要注意的是，绘制的时候四个顶点必须形成空间同平面的平行四边形
--- 否则会出现失真的现象，在这个教程里面也会演示出来

function tutorial.render_by_vertex()
	load_resource()
	
	local host = create_host(
			function(self)
				self.layer = LAYER_BG + 50 -- 图层，可以修改
			end,
			function(self)
				-- 什么都不做
			end,
			function(self)
				local k = 0.5 + 0.5 * lstg.sin(self.timer * 8)
				-- 在这里，我们可以指定每个顶点的颜色
				-- 顶点（x坐标，y坐标，z坐标，u坐标，v坐标，颜色）
				-- 参数（纹理名，混合模式，顶点1，顶点2，顶点3，顶点4）
				lstg.RenderTexture("tex_pillar", "mul+alpha",
						{ -160,  160, 0.5,  0,   0, lstg.Color(255,     255, k * 255, k * 255) },
						{ -120,  160, 0.5, 32,   0, lstg.Color(255, k * 255,     255, k * 255) },
						{ -120, -160, 0.5, 32, 256, lstg.Color(255, k * 255, k * 255,     255) },
						{ -160, -160, 0.5,  0, 256, lstg.Color(255,     255,     255,     255) })
				-- 失真的情况
				lstg.RenderTexture("tex_pillar", "mul+alpha",
						{ -100,  16, 0.5,  0,   0, lstg.Color(255,     255, k * 255, k * 255) },
						{  160,  64, 0.5, 32,   0, lstg.Color(255, k * 255,     255, k * 255) },
						{   50, -32, 0.5, 32, 256, lstg.Color(255, k * 255, k * 255,     255) },
						{  -80, -16, 0.5,  0, 256, lstg.Color(255,     255,     255,     255) })
			end)
	
	infinite_wait()
end

----------------------------------------
--- 教程 指定每个顶点的颜色

--- 在不使用 lstg.RenderTexture 的情况下，也能修改图片每个顶点的颜色
--- 这时候就得介绍 lstg.SetImageState 的第二种调用方式

function tutorial.change_4_vertex_color()
	load_resource()
	
	local host = create_host(
			function(self)
				self.layer = LAYER_BG + 50 -- 图层，可以修改
			end,
			function(self)
				-- 什么都不做
			end,
			function(self)
				-- 参数（图片名，混合模式，颜色1，颜色2，颜色3，颜色4）
				lstg.SetImageState("img_pillar", "mul+alpha",
						lstg.Color(255, 255,   0,   0),
						lstg.Color(255,   0, 255,   0),
						lstg.Color(255,   0,   0, 255),
						lstg.Color(255, 255, 255, 255))
				-- 参数（图片名，x坐标，y坐标，朝向，横向缩放，纵向缩放）
				lstg.Render("img_pillar", 0, 0, 0, 8, 1) -- 然后绘制
			end)
	
	infinite_wait()
end

----------------------------------------
--- 教程 绘制动画

--- 动画是和游戏对象绑定的，随游戏对象一起更新
--- 功能非常简陋，一般做复杂动画的时候会自己用图片来实现

function tutorial.render_animation()
	load_resource()
	
	local host = create_host(
			function(self)
				self.layer = LAYER_BG + 50 -- 图层，可以修改
				
				-- DefaultRenderFunc 会用到的游戏对象属性
				self.img = "ani_pillar"
				self.x = 0 -- x 坐标
				self.y = 0 -- y 坐标
				self.rot = 30 -- 朝向，角度制，逆时针为正方向
				self.hscale = 3 -- 横向缩放
				self.vscale = 3 -- 纵向缩放
				
				-- 其他属性
				self.omiga = 0 -- 自转
			end,
			function(self)
				-- 什么都不做
			end,
			function(self)
				local k = 0.5 + 0.5 * lstg.sin(self.timer * 8)
				-- 参数（游戏对象，混合模式，不透明度，红色，绿色，蓝色）
				lstg.SetImgState(self, "mul+alpha", 255, 255, k * 255, k * 255) -- 先设置
				-- 参数（游戏对象）
				lstg.DefaultRenderFunc(self) -- 然后绘制，这个是游戏对象的默认绘制函数
			end)
	
	infinite_wait()
end

----------------------------------------
--- 教程 绘制粒子特效

--- 粒子特效相比起图片和动画来说，更加复杂
--- 粒子发射器是和游戏对象绑定的，随游戏对象一起更新
--- 对性能要求比较高，在粒子特效多的时候容易掉帧
--- 有一些额外的方法可以控制粒子发射器

function tutorial.render_particle()
	load_resource()
	
	local host = create_host(
			function(self)
				self.layer = LAYER_BG + 50 -- 图层，可以修改
				
				-- DefaultRenderFunc 会用到的游戏对象属性
				self.img = "psi_death"
				self.x = 0 -- x 坐标
				self.y = 0 -- y 坐标
				self.rot = 0 -- 朝向，角度制，逆时针为正方向
				self.hscale = 0.25 -- 横向缩放
				self.vscale = 0.25 -- 纵向缩放
				
				-- 其他属性
				self.omiga = 2 -- 自转
				
				task.New(self, function()
					while true do
						-- 启动绑定在游戏对象上的粒子发射器
						-- 参数（游戏对象）
						lstg.ParticleFire(self)
						-- 修改绑定在游戏对象上的粒子发射器的每秒最大发射粒子数
						-- 参数（游戏对象，每秒最大发射粒子数）
						lstg.ParticleSetEmission(self, 500)
						task.Wait(120)
						-- 获取绑定在游戏对象上的粒子发射器的每秒最大发射粒子数
						-- 参数（游戏对象）
						-- 返回（每秒最大发射粒子数）
						if lstg.ParticleGetEmission(self) ~= 500 then
							error("奇怪了，为什么数目不匹配")
						end
						-- 修改绑定在游戏对象上的粒子发射器的每秒最大发射粒子数
						-- 参数（游戏对象，每秒最大发射粒子数）
						lstg.ParticleSetEmission(self, 50)
						task.Wait(120)
						-- 停止绑定在游戏对象上的粒子发射器
						-- 参数（游戏对象）
						lstg.ParticleStop(self)
						-- 获取绑定在游戏对象上的粒子发射器的活跃粒子数
						-- 参数（游戏对象）
						-- 返回（活跃粒子数）
						while lstg.ParticleGetn(self) > 0 do
							task.Wait(1)
						end
					end
				end)
			end,
			function(self)
				-- 什么都不做
			end,
			function(self)
				-- 参数（游戏对象）
				lstg.DefaultRenderFunc(self) -- 然后绘制，这个是游戏对象的默认绘制函数
			end)
	
	infinite_wait()
end

----------------------------------------
--- 教程 文字显示

--- 很多时候我们需要看到结果，lstg.Print 似乎不是一个最好的选择
--- 于是，我们可以通过显示文字来直观地看到一些结果
--- 文字绘制的时候，以 xy 坐标为基准，根据对齐方式绘制
--- 需要注意的是 lstg.RenderTTF 的缩放参数会自动乘以 0.5
--- 这是一个历史遗留问题，在使用的时候需要注意

function tutorial.render_text()
	load_resource()
	
	local host = create_host(
			function(self)
				self.layer = LAYER_BG + 50 -- 图层，可以修改
			end,
			function(self)
				-- 什么都不做
			end,
			function(self)
				-- 参数（HGE 纹理字体名，混合模式，颜色）
				lstg.SetFontState("fnt_menu", "mul+alpha", lstg.Color(255, 255, 255, 0))
				-- 参数（HGE 纹理字体名，要显示的文字，x坐标，y坐标，缩放，对齐方式）
				lstg.RenderText("fnt_menu", "Tutorial2020", -160, 100, 1, 0 + 4)
				-- 参数（TrueType 字体名，要显示的文字，左边，右边，底边，顶边，对齐方式，颜色，缩放）
				lstg.RenderTTF("ttf_unifont", "教程2020",
						-160, 100, -160, -100,
						0 + 0, lstg.Color(255, 255, 255, 255), 1)
				-- 一般情况下，填x坐标=左边=右边，y坐标=底边=顶边
				-- 参数（TrueType 字体名，要显示的文字，左边，右边，底边，顶边，对齐方式，颜色，缩放）
				lstg.RenderTTF("ttf_unifont", string.format("我的计时器：%d", self.timer),
						-160, -160, 0, 0,
						0 + 4, lstg.Color(255, 255, 0, 0), 1)
			end)
	
	infinite_wait()
end

----------------------------------------
--- 教程 平行投影矩阵和视口设置

--- 在前面的教程中，我们一直是以默认的坐标系进行绘制的
--- 这个坐标系的设置方式为 SetViewMode("world")
--- 此外还有一个默认的 UI 坐标系，可以通过 SetViewMode("ui") 设置
--- 如果我们需要定义自己的坐标系，在平面上绘制，我们可以使用以下的函数
--- lstg.SetFog、lstg.SetViewport、lstg.SetOrtho

function tutorial.set_orthogonal()
	load_resource()
	
	local host = create_host(
			function(self)
				self.layer = LAYER_TOP + 50 -- 图层，可以修改
			end,
			function(self)
				-- 什么都不做
			end,
			function(self)
				-- 关闭顶点雾功能
				-- 参数（）
				lstg.SetFog()
				-- 设置可绘制区域，也就是视口，在 LuaSTG 中视口的各个参数以像素为单位，并以左下角为原点
				-- 比如游戏窗口大小是 800x600，想设置以一半大小在正中央绘制，就填写 200, 600, 150, 450
				-- 参数（左边）
				lstg.SetViewport(16, 16 + 200, 32, 32 + 200)
				-- 设置可绘制区域内的平行投影坐标系
				-- 参数（x轴最小值，x轴最大值，y轴最小值，y轴最大值）
				lstg.SetOrtho(0, 100, -50, 0)
				-- 设置图片、动画、粒子特效、字体、曲线激光绘制时的全局缩放
				-- 对于图片来说，就是图片的每个像素被对应到坐标系内的几个单位
				-- 比如一张 64x64 像素的图片，在 1 的绘制缩放和 0.5 的全局缩放下绘制，大小只有 32x32（不是像素单位，是坐标系单位）
				-- 参数（缩放）
				lstg.SetImageScale(1)
				
				-- 注意观察下面绘制时填写的顶点坐标和上面 lstg.SetOrtho 的参数
				
				-- 参数（图片名，混合模式，颜色）
				lstg.SetImageState("img_pillar", "mul+alpha", lstg.Color(255, 255, 255, 255)) -- 先设置
				-- 参数（图片名，x坐标1，y坐标1，z坐标1，x坐标2，y坐标2，z坐标2，x坐标3，y坐标3，z坐标3，x坐标4，y坐标4，z坐标4）
				lstg.Render4V("img_pillar",
						  0,   0, 0.5,
						100,   0, 0.5,
						100, -50, 0.5,
						  0, -50, 0.5)
				
				SetViewMode("world") -- 当然，要记得恢复回来
			end)
	
	infinite_wait()
end

----------------------------------------
--- 教程 透视投影矩阵和视口设置

--- 通过默认的 3d 坐标系，我们可以绘制简单的3D画面
--- 通过调用 SetViewMode("3d") 来切换到 3d 坐标系
--- 如果我们需要定义自己的坐标系，在3D空间上绘制，我们可以使用以下的函数
--- lstg.SetFog、lstg.SetViewport、lstg.SetPerspective

function tutorial.set_perspective()
	load_resource()
	
	local host = create_host(
			function(self)
				self.layer = LAYER_TOP + 50 -- 图层，可以修改
			end,
			function(self)
				-- 什么都不做
			end,
			function(self)
				-- 关闭顶点雾功能
				-- 参数（）
				lstg.SetFog()
				-- 设置可绘制区域，也就是视口，在 LuaSTG 中视口的各个参数以像素为单位，并以左下角为原点
				-- 比如游戏窗口大小是 800x600，想设置以一半大小在正中央绘制，就填写 200, 600, 150, 450
				-- 参数（左边）
				lstg.SetViewport(16, 16 + 400, 32, 32 + 200)
				-- 设置可绘制区域内的平行投影坐标系
				-- 参数（观察者x坐标，观察者y坐标，观察者z坐标，
				--         观察的x坐标，观察的y坐标，观察的z坐标，
				--         观察者的头顶朝向的向量x分量， 观察者的头顶朝向的向量y分量， 观察者的头顶朝向的向量z分量，
				--         弧度制视角，
				--         画面宽高比，
				--         近平面，远平面）
				lstg.SetPerspective(0, 0, 0, -- 在原点
						0, 0, 1, -- 看向z轴正方向
						0, 1, 0, -- 头顶朝向的向量，朝向y轴正方向
						math.rad(90), -- 90度的视角
						400 / 200, -- 宽高比为 2:1，根据视口填写这个参数
						1, 1000) -- 离观察者 1 到 1000 坐标系单位的面都会被绘制
				-- 虽然不需要，但是还是可以设置一下
				-- 一般在3D空间中绘制，用的是 lstg.Render4V 和 lstg.RenderTexture
				-- 这两个方法不受全局缩放影响，因为是开发者自己来指定顶点坐标
				-- 参数（缩放）
				lstg.SetImageScale(1)
				
				-- 参数（颜色）
				lstg.RenderClear(lstg.Color(255, 128, 128, 128)) -- 覆盖可绘制区域
				
				-- 参数（图片名，混合模式，颜色）
				lstg.SetImageState("img_pillar", "mul+alpha", lstg.Color(255, 255, 255, 255)) -- 先设置
				-- 参数（图片名，x坐标1，y坐标1，z坐标1，x坐标2，y坐标2，z坐标2，x坐标3，y坐标3，z坐标3，x坐标4，y坐标4，z坐标4）
				lstg.Render4V("img_pillar",
						-0.5,   0.5, 2,
						0.5,   0.5, 2,
						0.5, -0.5, 1,
						-0.5, -0.5, 1)
				
				SetViewMode("world") -- 当然，要记得恢复回来
			end)
	
	infinite_wait()
end

----------------------------------------
--- 教程 雾效果

--- lstg.SetFog 能让我们使用简单的雾效果

function tutorial.set_fog()
	load_resource()
	
	local host = create_host(
			function(self)
				self.layer = LAYER_TOP + 50 -- 图层，可以修改
			end,
			function(self)
				-- 什么都不做
			end,
			function(self)
				-- 开启顶点雾功能
				-- 参数（雾起始距离，雾最大距离，颜色）
				lstg.SetFog(2, 4, lstg.Color(255, 128, 128, 128))
				-- 设置可绘制区域，也就是视口，在 LuaSTG 中视口的各个参数以像素为单位，并以左下角为原点
				-- 比如游戏窗口大小是 800x600，想设置以一半大小在正中央绘制，就填写 200, 600, 150, 450
				-- 参数（左边）
				lstg.SetViewport(16, 16 + 400, 32, 32 + 200)
				-- 设置可绘制区域内的平行投影坐标系
				-- 参数（观察者x坐标，观察者y坐标，观察者z坐标，
				--         观察的x坐标，观察的y坐标，观察的z坐标，
				--         观察者的头顶朝向的向量x分量， 观察者的头顶朝向的向量y分量， 观察者的头顶朝向的向量z分量，
				--         弧度制视角，
				--         画面宽高比，
				--         近平面，远平面）
				lstg.SetPerspective(0, 0, 0, -- 在原点
						0, 0, 1, -- 看向z轴正方向
						0, 1, 0, -- 头顶朝向的向量，朝向y轴正方向
						math.rad(90), -- 90度的视角
						400 / 200, -- 宽高比为 2:1，根据视口填写这个参数
						1, 1000) -- 离观察者 1 到 1000 坐标系单位的面都会被绘制
				-- 虽然不需要，但是还是可以设置一下
				-- 一般在3D空间中绘制，用的是 lstg.Render4V 和 lstg.RenderTexture
				-- 这两个方法不受全局缩放影响，因为是开发者自己来指定顶点坐标
				-- 参数（缩放）
				lstg.SetImageScale(1)
				
				-- 参数（颜色）
				lstg.RenderClear(lstg.Color(255, 128, 128, 128)) -- 覆盖可绘制区域
				
				-- 参数（图片名，混合模式，颜色）
				lstg.SetImageState("img_pillar", "mul+alpha", lstg.Color(255, 255, 255, 255)) -- 先设置
				-- 参数（图片名，x坐标1，y坐标1，z坐标1，x坐标2，y坐标2，z坐标2，x坐标3，y坐标3，z坐标3，x坐标4，y坐标4，z坐标4）
				lstg.Render4V("img_pillar",
						-2,   0, 10,
						2,   0, 10,
						2, -1, 1,
						-2, -1, 1)
				
				SetViewMode("world") -- 当然，要记得恢复回来
			end)
	
	infinite_wait()
end

----------------------------------------
--- 教程 绘制到纹理和使用后处理特效

function tutorial.rendertarget_and_effect()
	load_resource()
	
	-- 白色背景
	local bg = create_host(
			function(self)
				self.layer = LAYER_BG -- 图层，可以修改
			end,
			function(self)
				-- 什么都不做
			end,
			function(self)
				-- 参数（颜色）
				lstg.RenderClear(lstg.Color(255, 255, 255, 255))
			end)
	
	-- 从图层 LAYER_BG - 50 到 LAYER_TOP + 50 的游戏对象都会绘制到 tex_buffer 上
	local host1 = create_host(
			function(self)
				self.layer = LAYER_BG - 50 -- 图层，可以修改
			end,
			function(self)
				-- 什么都不做
			end,
			function(self)
				-- 设置离屏渲染的纹理，纹理必须是通过 lstg.CreateRenderTarget 创建的纹理
				-- 参数（纹理名）
				lstg.PushRenderTarget("tex_buffer")
				-- 每次使用的时候记得要清空一次
				-- 参数（颜色）
				lstg.RenderClear(lstg.Color(255, 0, 0, 0))
			end)
	local host2 = create_host(
			function(self)
				self.layer = LAYER_TOP + 50 -- 图层，可以修改
			end,
			function(self)
				-- 什么都不做
			end,
			function(self)
				-- 取消使用离屏渲染的纹理
				-- 参数（）
				lstg.PopRenderTarget()
				
				-- 为了方便观察，先把背景清空成别的颜色，这一步不是必要的
				-- 参数（颜色）
				lstg.RenderClear(lstg.Color(255, 128, 128, 128))
				
				-- 应用后处理效果
				-- 参数表是一个 lua 的 table 类型
				-- 参数（纹理名，后处理特效名，混合模式，参数表）
				lstg.PostEffect("tex_buffer", "fx_effect", "mul+alpha", {
					color = lstg.Color(255, 255, 255, 255),
					number = 0.75 + 0.25 * lstg.sin(self.timer * 4),
					texture = "tex_pillar",
				})
			end)
	
	infinite_wait()
end
